//  Copyright 2004 The Apache Software Foundation
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.apache.hivemind.sdl;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStream;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.hivemind.HiveMind;
import org.apache.hivemind.Resource;
import org.apache.hivemind.sdl.parser.SimpleDataLanguageParser;
import org.apache.hivemind.sdl.parser.TokenMgrError;
import org.xml.sax.ContentHandler;
import org.xml.sax.Locator;
import org.xml.sax.SAXParseException;

/**
 * Wrapper around {@link org.apache.hivemind.sdl.parser.SimpleDataLanguageParser}
 * (which is generated code). Parses resources that contained Simple Data Language - a language
 * inspired by XML, but rationalized for storing hiearachical data (a frequent application of XML,
 * even though XML is better suited for rich document markup than storing of data).  SDL
 * is isomorphic to a subset of XML: It has equivalents for XML elements, XML attributes,
 * and parsable character data (text within an element), but does not support processing
 * instructions, namespaces, internal or external entities, doctype, etc.
 *
 * @author Howard Lewis Ship
 */
public class SDLResourceParser
{
    private static final Log LOG = LogFactory.getLog(SDLResourceParser.class);

    /**
     * Parses the resource, passing SAX parse events to the handler. SDL does
     * not (at this time) support namespaces or other constructs of XML, so only
     * the following methods of ContentHandler will be invoked:
     * <ul>
     * <li>{@link ContentHandler#setDocumentLocator(org.xml.sax.Locator)} - before any other calls
     * <ii>{@link ContentHandler#startDocument()}
     * <li>{@link ContentHandler#startElement(java.lang.String, java.lang.String, java.lang.String, org.xml.sax.Attributes)
     * 		- passing only the local name (namespace URL and qname will be null)
     * <li>{@link ContentHandler#characters(char[], int, int) for literal (and extended literal) strings
     * <li>{@link ContentHandler#endElement(java.lang.String, java.lang.String, java.lang.String) - again, with the namespace
     * parameters passed as null
     * <li>{@link ContentHandler#endDocument() - last
     * </ul>
     * 
     * <p>
     * At some future point, SDL may support more XML-like things, including names spaces.
     */
    public void parseResource(Resource resource, ContentHandler handler)
        throws SAXParseException, IOException
    {
        BufferedInputStream bis = null;
        SimpleDataLanguageParser parser = null;
        Locator locator = null;

        try
        {
            InputStream rawStream = resource.getResourceURL().openStream();

            bis = new BufferedInputStream(rawStream);

            parser = new SimpleDataLanguageParser(rawStream);
            locator = parser.getLocator();

            parser.parse(handler);

            bis.close();
            bis = null;
        }
        catch (TokenMgrError ex)
        {
            throw new SAXParseException(SDLMessages.parseError(resource, ex), locator);
        }
        catch (Exception ex)
        {
            throw new SAXParseException(SDLMessages.parseError(resource, ex), locator, ex);
        }
        finally
        {
            close(resource, bis);
        }
    }

    private void close(Resource resource, InputStream stream)
    {
        try
        {
            if (stream != null)
                stream.close();
        }
        catch (IOException ex)
        {
            LOG.error(SDLMessages.unableToCloseStream(resource, ex), ex);
        }
    }
}
