//  Copyright 2004 The Apache Software Foundation
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.apache.hivemind.service.impl;

import org.apache.commons.logging.Log;
import org.apache.hivemind.service.ClassFabUtils;

/**
 * Base class used to dynamically build logging interceptors.
 * {@link org.apache.hivemind.service.impl.LoggingInterceptorFactory}
 * will create a subclass dynamically.
 *
 * @author Howard Lewis Ship
 */
public abstract class AbstractLoggingInterceptor
{
    private static final int BUFFER_SIZE = 100;

    private Log _log;

    protected AbstractLoggingInterceptor(Log log)
    {
        _log = log;
    }

    protected void _logEntry(String methodName, Object[] args)
    {
        StringBuffer buffer = new StringBuffer(BUFFER_SIZE);

        buffer.append("BEGIN ");
        buffer.append(methodName);
        buffer.append("(");

        int count = (args == null) ? 0 : args.length;

        for (int i = 0; i < count; i++)
        {
            Object arg = args[i];

            if (i > 0)
                buffer.append(", ");

            convert(buffer, arg);
        }

        buffer.append(")");

        _log.debug(buffer.toString());
    }

    private void convert(StringBuffer buffer, Object input)
    {
        if (input == null)
        {
            buffer.append("<null>");
            return;
        }

        // Primitive types, and non-object arrays
        // use toString().  Less than ideal for int[], etc., but
        // that's a lot of work for a rare case.

        if (!(input instanceof Object[]))
        {
            buffer.append(input.toString());
            return;
        }

        buffer.append("(");
        buffer.append(ClassFabUtils.getJavaClassName(input.getClass()));
        buffer.append("){");

        Object[] array = (Object[]) input;
        int count = array.length;

        for (int i = 0; i < count; i++)
        {
            if (i > 0)
                buffer.append(", ");

            // We use convert() again, because it could be a multi-dimensional array
            // (god help us) where each element must be converted.
            convert(buffer, array[i]);
        }

        buffer.append("}");
    }

    protected void _logExit(String methodName, Object result)
    {
        StringBuffer buffer = new StringBuffer(BUFFER_SIZE);

        buffer.append("END ");
        buffer.append(methodName);
        buffer.append("() [");

        convert(buffer, result);

        buffer.append("]");

        _log.debug(buffer.toString());
    }

    protected void _logVoidExit(String methodName)
    {
        StringBuffer buffer = new StringBuffer(BUFFER_SIZE);

        buffer.append("END ");
        buffer.append(methodName);
        buffer.append("()");

        _log.debug(buffer.toString());
    }

    protected void _logException(String methodName, Throwable t)
    {
        StringBuffer buffer = new StringBuffer(BUFFER_SIZE);

        buffer.append("EXCEPTION ");
        buffer.append(methodName);
        buffer.append("() -- ");

        buffer.append(t.getClass().getName());

        _log.debug(buffer.toString(), t);
    }

    protected boolean _isDebugEnabled()
    {
        return _log.isDebugEnabled();
    }

}