//  Copyright 2004 The Apache Software Foundation
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.apache.hivemind.service.impl;

import java.lang.reflect.Method;
import java.util.Iterator;
import java.util.List;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.hivemind.ApplicationRuntimeException;
import org.apache.hivemind.ClassResolver;
import org.apache.hivemind.ErrorHandler;
import org.apache.hivemind.HiveMind;
import org.apache.hivemind.Location;
import org.apache.hivemind.ServiceImplementationFactory;
import org.apache.hivemind.internal.Module;
import org.apache.hivemind.service.EventLinker;
import org.apache.hivemind.util.ConstructorUtils;
import org.apache.hivemind.util.PropertyUtils;

/**
 * Implementation of {@link org.apache.hivemind.ServiceImplementationFactory}
 * that can instantiate an object and then configure its properties.
 * 
 * <p>
 * Some thought has been given to using bytecode generation to create properties
 * for messages, extension point id, and so forth.  This is being avoided because it
 * undermines the ability to test service implemenations as POJOs, outside the
 * framework of HiveMind.
 * 
 * <p>
 * Instead the service is configured by means of the implementation's
 * constructor and setter methods.
 *
 * @author Howard Lewis Ship
 */
public class BuilderFactory implements ServiceImplementationFactory
{
    private static final String POINT_ID = "hivemind.BuilderFactory";

    public Object createCoreServiceImplementation(
        String serviceId,
        Class serviceInterface,
        Module invokingModule,
        List parameters)
    {
        HiveMind.checkFactoryParameterCount(POINT_ID, parameters, 1);

        Log log = LogFactory.getLog(serviceId);
        ErrorHandler errorHandler = invokingModule.getErrorHandler();

        BuilderParameter parameter = (BuilderParameter) parameters.get(0);

        Object result = instantiateCoreServiceInstance(serviceId, invokingModule, parameter);

        setProperties(serviceId, invokingModule, parameter, result, log);

        registerForEvents(log, errorHandler, parameter, result);

        invokeInitializer(parameter, result, serviceId, log, invokingModule);

        return result;
    }

    private void registerForEvents(
        Log log,
        ErrorHandler eventHandler,
        BuilderParameter parameter,
        Object result)
    {
        List eventRegistrations = parameter.getEventRegistrations();

        if (eventRegistrations.isEmpty())
            return;

        EventLinker linker = new EventLinkerImpl(log, eventHandler);
        Iterator i = eventRegistrations.iterator();
        while (i.hasNext())
        {
            EventRegistration er = (EventRegistration) i.next();

            linker.addEventListener(
                er.getProducer(),
                er.getEventSetName(),
                result,
                er.getLocation());
        }
    }

    private void setProperties(
        String serviceId,
        Module invokingModule,
        BuilderParameter parameter,
        Object target,
        Log log)
    {
        List properties = parameter.getProperties();
        int count = properties.size();

        for (int i = 0; i < count; i++)
        {
            BuilderFacet facet = (BuilderFacet) properties.get(i);
            String propertyName = facet.getPropertyName();

			// Facets that can autowire, should.
			
            facet.autowire(target, serviceId, invokingModule);

            // There will be a facet for log, messages, service-id, etc. even if no
            // property name is specified, so we skip it here.  In many cases, those
            // facets will have just done an autowire.

            if (propertyName == null)
                continue;

            Class targetType = PropertyUtils.getPropertyType(target, propertyName);

            Object value = facet.getFacetValue(serviceId, invokingModule, targetType);

            Location location = HiveMind.getLocation(facet);

            HiveMind.setLocation(value, location);

            try
            {
                PropertyUtils.write(target, propertyName, value);
            }
            catch (Exception ex)
            {
                throw new ApplicationRuntimeException(ex.getMessage(), target, location, ex);
            }
        }
    }

    private Object instantiateCoreServiceInstance(
        String serviceId,
        Module invokingModule,
        BuilderParameter parameter)
    {

        ClassResolver resolver = invokingModule.getClassResolver();
        Class serviceClass = resolver.findClass(parameter.getClassName());

        Object[] constructorParameters =
            buildConstructorParameters(parameter, invokingModule, serviceId);

        return ConstructorUtils.invokeConstructor(serviceClass, constructorParameters, parameter);
    }

    private Object[] buildConstructorParameters(
        BuilderParameter parameter,
        Module invokingModule,
        String serviceId)
    {
        List parameters = parameter.getParameters();
        int count = parameters.size();

        if (count == 0)
            return null;

        Object[] result = new Object[count];

        for (int i = 0; i < count; i++)
        {
            BuilderFacet facet = (BuilderFacet) parameters.get(i);

            result[i] = facet.getFacetValue(serviceId, invokingModule, Object.class);

            HiveMind.setLocation(result[i], HiveMind.getLocation(facet));
        }

        return result;
    }

    private void invokeInitializer(
        BuilderParameter parameters,
        Object service,
        String serviceId,
        Log log,
        Module module)
    {
        String methodName = parameters.getInitializeMethod();

        if (HiveMind.isBlank(methodName))
            return;

        try
        {
            Class serviceClass = service.getClass();
            Method m = serviceClass.getMethod(methodName, null);

            m.invoke(service, null);
        }
        catch (Exception ex)
        {
            module.getErrorHandler().error(
                log,
                ServiceMessages.unableToInitializeService(
                    serviceId,
                    methodName,
                    service.getClass(),
                    ex),
                parameters.getLocation(),
                ex);
        }
    }
}
