//  Copyright 2004 The Apache Software Foundation
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package hivemind.test.sdl;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import org.apache.hivemind.Attribute;
import org.apache.hivemind.Element;
import org.apache.hivemind.Location;
import org.apache.hivemind.Resource;
import org.apache.hivemind.impl.AttributeImpl;
import org.apache.hivemind.impl.ElementImpl;
import org.apache.hivemind.impl.LocationImpl;
import org.apache.hivemind.sdl.SDLResourceParser;
import org.xml.sax.Attributes;
import org.xml.sax.ContentHandler;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

/**
 * Used to test the Simple Data Language parser. Assembles
 * a tree of {@link org.apache.hivemind.Element}s from a document. Currently,
 * this is used for testing ... but its pretty useful, and may move into
 * the main framework (perhaps as a service of some kind) in the future.
 *
 * @author Howard Lewis Ship
 */
public class ElementAssembler implements ContentHandler
{
    private Element _root = null;

    private class StackItem
    {
        private ElementImpl _element;
        private StringBuffer _buffer = new StringBuffer();

        StackItem(ElementImpl element)
        {
            _element = element;
        }

        void addChild(Element e)
        {
            _element.addElement(e);
        }

        void addContent(String content)
        {
            _buffer.append(content);
        }

        void end()
        {
            if (_buffer.length() > 0)
                _element.setContent(_buffer.toString());
        }
    }

    /** Stack of StackItem **/

    private List _stack = new ArrayList();

    private Resource _resource;
    private Locator _locator;

    private StackItem peek()
    {
        int count = _stack.size();

        return (StackItem) _stack.get(count - 1);
    }

    private void push(ElementImpl e)
    {
        _stack.add(new StackItem(e));
    }

    private void pop()
    {
        int count = _stack.size();

        StackItem s = (StackItem) _stack.remove(count - 1);

        s.end();
    }

    public void startElement(
        String namespaceURI,
        String localName,
        String qName,
        Attributes attributes)
        throws SAXException
    {
        ElementImpl element = constructElement(localName, attributes);

        push(element);
    }

    public void characters(char[] ch, int start, int length) throws SAXException
    {
        String content = new String(ch, start, length);

        peek().addContent(content);
    }

    public void endElement(String namespaceURI, String localName, String qName) throws SAXException
    {
        pop();
    }

    public void endPrefixMapping(String prefix) throws SAXException
    {

    }

    public void ignorableWhitespace(char[] ch, int start, int length) throws SAXException
    {

    }

    public void processingInstruction(String target, String data) throws SAXException
    {

    }

    public void setDocumentLocator(Locator locator)
    {
        _locator = locator;
    }

    public void skippedEntity(String name) throws SAXException
    {

    }

    public void startDocument() throws SAXException
    {

    }

    public void endDocument() throws SAXException
    {

    }

    public void startPrefixMapping(String prefix, String uri) throws SAXException
    {

    }

    private ElementImpl constructElement(String name, Attributes attributes)
    {
        ElementImpl result = new ElementImpl();
        result.setElementName(name);
        result.setLocation(getLocation());

        int count = attributes.getLength();
        for (int i = 0; i < count; i++)
        {
            Attribute a = new AttributeImpl(attributes.getLocalName(i), attributes.getValue(i));

            result.addAttribute(a);
        }

        if (_root == null)
            _root = result;
        else
            peek().addChild(result);

        return result;
    }

    private Location getLocation()
    {
        return new LocationImpl(_resource, _locator.getLineNumber(), _locator.getColumnNumber());
    }

    public Element parse(Resource resource) throws SAXParseException, IOException
    {
        SDLResourceParser p = new SDLResourceParser();

        p.parseResource(resource, this);

        return _root;
    }
}
